@@ -10,15 +10,18 @@ import string
 import sys
 import uuid
 import warnings
+from collections import OrderedDict
 from fnmatch import fnmatchcase
 from subprocess import list2cmdline
 
 import pkg_resources
 import pluggy
 import py
+import toml
 
 import tox
-from tox.interpreters import Interpreters
+from tox.constants import INFO
+from tox.interpreters import Interpreters, NoInterpreterInfo
 
 hookimpl = tox.hookimpl
 """DEPRECATED - REMOVE - this is left for compatibility with plugins importing this from here.
@@ -95,7 +98,7 @@ class Parser:
         assert hasattr(obj, "postprocess")
         self._testenv_attr.append(obj)
 
-    def _parse_args(self, args):
+    def parse_cli(self, args):
         return self.argparser.parse_args(args)
 
     def _format_help(self):
@@ -136,7 +139,8 @@ class DepOption:
                         name = "{}{}".format(option, name[len(option) :].strip())
                 # in case of a long option, we add an equal sign
                 for option in tox.PIP.INSTALL_LONG_OPTIONS_ARGUMENT:
-                    if name.startswith(option + " "):
+                    name_start = "{} ".format(option)
+                    if name.startswith(name_start):
                         name = "{}={}".format(option, name[len(option) :].strip())
             name = self._replace_forced_dep(name, config)
             deps.append(DepConfig(name, ixserver))
@@ -213,53 +217,70 @@ def parseconfig(args, plugins=()):
     :raise SystemExit: toxinit file is not found
     """
     pm = get_plugin_manager(plugins)
-    # prepare command line options
-    parser = Parser()
-    pm.hook.tox_addoption(parser=parser)
-    # parse command line options
-    option = parser._parse_args(args)
-    interpreters = Interpreters(hook=pm.hook)
-    config = Config(pluginmanager=pm, option=option, interpreters=interpreters)
-    config._parser = parser
-    config._testenv_attr = parser._testenv_attr
-    if config.option.version:
-        print(get_version_info(pm))
-        raise SystemExit(0)
-    # parse ini file
-    basename = config.option.configfile
-    if os.path.isfile(basename):
-        inipath = py.path.local(basename)
-    elif os.path.isdir(basename):
-        # Assume 'tox.ini' filename if directory was passed
-        inipath = py.path.local(os.path.join(basename, "tox.ini"))
+    config, option = parse_cli(args, pm)
+
+    for config_file in propose_configs(option.configfile):
+        config_type = config_file.basename
+
+        content = None
+        if config_type == "pyproject.toml":
+            toml_content = get_py_project_toml(config_file)
+            try:
+                content = toml_content["tool"]["tox"]["legacy_tox_ini"]
+            except KeyError:
+                continue
+        ParseIni(config, config_file, content)
+        pm.hook.tox_configure(config=config)  # post process config object
+        break
     else:
-        for path in py.path.local().parts(reverse=True):
-            inipath = path.join(basename)
-            if inipath.check():
-                break
-        else:
-            inipath = py.path.local().join("setup.cfg")
-            if not inipath.check():
-                helpoptions = option.help or option.helpini
-                feedback("toxini file {!r} not found".format(basename), sysexit=not helpoptions)
-                if helpoptions:
-                    return config
-
-    try:
-        parseini(config, inipath)
-    except tox.exception.InterpreterNotFound:
-        exn = sys.exc_info()[1]
-        # Use stdout to match test expectations
-        print("ERROR: {}".format(exn))
-
-    # post process config object
-    pm.hook.tox_configure(config=config)
-
+        msg = "tox config file (either {}) not found"
+        candidates = ", ".join(INFO.CONFIG_CANDIDATES)
+        feedback(msg.format(candidates), sysexit=not (option.help or option.helpini))
     return config
 
 
+def get_py_project_toml(path):
+    with open(str(path)) as file_handler:
+        config_data = toml.load(file_handler)
+        return config_data
+
+
+def propose_configs(cli_config_file):
+    from_folder = py.path.local()
+    if cli_config_file is not None:
+        if os.path.isfile(cli_config_file):
+            yield py.path.local(cli_config_file)
+            return
+        if os.path.isdir(cli_config_file):
+            from_folder = py.path.local(cli_config_file)
+        else:
+            print(
+                "ERROR: {} is neither file or directory".format(cli_config_file), file=sys.stderr
+            )
+            return
+    for basename in INFO.CONFIG_CANDIDATES:
+        if from_folder.join(basename).isfile():
+            yield from_folder.join(basename)
+        for path in from_folder.parts(reverse=True):
+            ini_path = path.join(basename)
+            if ini_path.check():
+                yield ini_path
+
+
+def parse_cli(args, pm):
+    parser = Parser()
+    pm.hook.tox_addoption(parser=parser)
+    option = parser.parse_cli(args)
+    if option.version:
+        print(get_version_info(pm))
+        raise SystemExit(0)
+    interpreters = Interpreters(hook=pm.hook)
+    config = Config(pluginmanager=pm, option=option, interpreters=interpreters, parser=parser)
+    return config, option
+
+
 def feedback(msg, sysexit=False):
-    print("ERROR: " + msg, file=sys.stderr)
+    print("ERROR: {}".format(msg), file=sys.stderr)
     if sysexit:
         raise SystemExit(1)
 
@@ -341,8 +362,9 @@ def tox_addoption(parser):
         action="count",
         dest="verbose_level",
         default=0,
-        help="increase verbosity of reporting output. -vv mode turns off "
-        "output redirection for package installation",
+        help="increase verbosity of reporting output."
+        "-vv mode turns off output redirection for package installation, "
+        "above level two verbosity flags are passed through to pip (with two less level)",
     )
     parser.add_argument(
         "-q",
@@ -373,7 +395,7 @@ def tox_addoption(parser):
     parser.add_argument(
         "-c",
         action="store",
-        default="tox.ini",
+        default=None,
         dest="configfile",
         help="config file name or directory with 'tox.ini' file.",
     )
@@ -473,11 +495,8 @@ def tox_addoption(parser):
     parser.add_argument(
         "--alwayscopy", action="store_true", help="override alwayscopy setting to True in all envs"
     )
-    parser.add_argument(
-        "--skip-missing-interpreters",
-        action="store_true",
-        help="don't fail tests for missing interpreters",
-    )
+
+    cli_skip_missing_interpreter(parser)
     parser.add_argument(
         "--workdir",
         action="store",
@@ -505,6 +524,9 @@ def tox_addoption(parser):
         config = testenv_config.config
         if "PYTHONHASHSEED" not in setenv and config.hashseed is not None:
             setenv["PYTHONHASHSEED"] = config.hashseed
+
+        setenv["TOX_ENV_NAME"] = str(testenv_config.envname)
+        setenv["TOX_ENV_DIR"] = str(testenv_config.envdir)
         return setenv
 
     parser.add_testenv_attribute(
@@ -515,35 +537,40 @@ def tox_addoption(parser):
     )
 
     def basepython_default(testenv_config, value):
-        """Configure a sane interpreter for the environment.
+        """either user set or proposed from the factor name
 
-        If the environment contains a default factor, this will always be the
-        interpreter associated with that factor overriding anything manually
-        set.
+        in both cases we check that the factor name implied python version and the resolved
+        python interpreter version match up; if they don't we warn, unless ignore base
+        python conflict is set in which case the factor name implied version if forced
         """
         for factor in testenv_config.factors:
-            match = tox.PYTHON.PY_FACTORS_RE.match(factor)
-            if match:
-                base_exe = tox.PYTHON.PY_FACTORS_MAP[match.group(1)]
-                version = ".".join(match.group(2) or "")
-                default = "{}{}".format(base_exe, version)
+            if factor in tox.PYTHON.DEFAULT_FACTORS:
+                implied_python = tox.PYTHON.DEFAULT_FACTORS[factor]
+                break
+        else:
+            implied_python, factor = None, None
 
-                if value is None or testenv_config.config.ignore_basepython_conflict:
-                    return default
+        if testenv_config.config.ignore_basepython_conflict and implied_python is not None:
+            return implied_python
 
-                if str(value) != default:
+        proposed_python = (implied_python or sys.executable) if value is None else str(value)
+        if implied_python is not None and implied_python != proposed_python:
+            testenv_config.basepython = proposed_python
+            implied_version = tox.PYTHON.PY_FACTORS_RE.match(factor).group(2)
+            python_info_for_proposed = testenv_config.python_info
+            if not isinstance(python_info_for_proposed, NoInterpreterInfo):
+                proposed_version = "".join(
+                    str(i) for i in python_info_for_proposed.version_info[0:2]
+                )
+                if implied_version != proposed_version:
                     # TODO(stephenfin): Raise an exception here in tox 4.0
                     warnings.warn(
-                        "Conflicting basepython for environment '{}'; resolve conflict "
-                        "or configure ignore_basepython_conflict".format(
-                            testenv_config.envname, str(value), default
+                        "conflicting basepython version (set {}, should be {}) for env '{}';"
+                        "resolve conflict or set ignore_basepython_conflict".format(
+                            proposed_version, implied_version, testenv_config.envname
                         )
                     )
-
-        if value is None:
-            return sys.executable
-
-        return str(value)
+        return proposed_python
 
     parser.add_testenv_attribute(
         name="basepython",
@@ -623,7 +650,7 @@ def tox_addoption(parser):
         # Flatten the list to deal with space-separated values.
         value = list(itertools.chain.from_iterable([x.split(" ") for x in value]))
 
-        passenv = {"PATH", "PIP_INDEX_URL", "LANG", "LANGUAGE", "LD_LIBRARY_PATH"}
+        passenv = {"PATH", "PIP_INDEX_URL", "LANG", "LANGUAGE", "LD_LIBRARY_PATH", "TOX_WORK_DIR"}
 
         # read in global passenv settings
         p = os.environ.get("TOX_TESTENV_PASSENV", None)
@@ -745,6 +772,20 @@ def tox_addoption(parser):
         help="each line specifies a test command and can use substitution.",
     )
 
+    parser.add_testenv_attribute(
+        name="commands_pre",
+        type="argvlist",
+        default="",
+        help="each line specifies a setup command action and can use substitution.",
+    )
+
+    parser.add_testenv_attribute(
+        name="commands_post",
+        type="argvlist",
+        default="",
+        help="each line specifies a teardown command and can use substitution.",
+    )
+
     parser.add_testenv_attribute(
         "ignore_outcome",
         type="bool",
@@ -760,16 +801,37 @@ def tox_addoption(parser):
     )
 
 
+def cli_skip_missing_interpreter(parser):
+    class SkipMissingInterpreterAction(argparse.Action):
+        def __call__(self, parser, namespace, values, option_string=None):
+            value = "true" if values is None else values
+            if value not in ("config", "true", "false"):
+                raise argparse.ArgumentTypeError("value must be config, true or false")
+            setattr(namespace, self.dest, value)
+
+    parser.add_argument(
+        "--skip-missing-interpreters",
+        default="config",
+        metavar="val",
+        nargs="?",
+        action=SkipMissingInterpreterAction,
+        help="don't fail tests for missing interpreters: {config,true,false} choice",
+    )
+
+
 class Config(object):
     """Global Tox config object."""
 
-    def __init__(self, pluginmanager, option, interpreters):
-        self.envconfigs = {}
+    def __init__(self, pluginmanager, option, interpreters, parser):
+        self.envconfigs = OrderedDict()
         """Mapping envname -> envconfig"""
         self.invocationcwd = py.path.local()
         self.interpreters = interpreters
         self.pluginmanager = pluginmanager
         self.option = option
+        self._parser = parser
+        self._testenv_attr = parser._testenv_attr
+
         """option namespace containing all parsed command line options"""
 
     @property
@@ -872,38 +934,36 @@ def make_hashseed():
     return str(random.randint(1, max_seed))
 
 
-class parseini:
-    def __init__(self, config, inipath):  # noqa
-        config.toxinipath = inipath
+class ParseIni(object):
+    def __init__(self, config, ini_path, ini_data):  # noqa
+        config.toxinipath = ini_path
         config.toxinidir = config.toxinipath.dirpath()
 
-        self._cfg = py.iniconfig.IniConfig(config.toxinipath)
+        self._cfg = py.iniconfig.IniConfig(config.toxinipath, ini_data)
         config._cfg = self._cfg
         self.config = config
 
-        if inipath.basename == "setup.cfg":
-            prefix = "tox"
-        else:
-            prefix = None
-        ctxname = getcontextname()
-        if ctxname == "jenkins":
+        prefix = "tox" if ini_path.basename == "setup.cfg" else None
+
+        context_name = getcontextname()
+        if context_name == "jenkins":
             reader = SectionReader(
                 "tox:jenkins", self._cfg, prefix=prefix, fallbacksections=["tox"]
             )
-            distshare_default = "{toxworkdir}/distshare"
-        elif not ctxname:
+            dist_share_default = "{toxworkdir}/distshare"
+        elif not context_name:
             reader = SectionReader("tox", self._cfg, prefix=prefix)
-            distshare_default = "{homedir}/.tox/distshare"
+            dist_share_default = "{homedir}/.tox/distshare"
         else:
             raise ValueError("invalid context")
 
         if config.option.hashseed is None:
-            hashseed = make_hashseed()
+            hash_seed = make_hashseed()
         elif config.option.hashseed == "noset":
-            hashseed = None
+            hash_seed = None
         else:
-            hashseed = config.option.hashseed
-        config.hashseed = hashseed
+            hash_seed = config.option.hashseed
+        config.hashseed = hash_seed
 
         reader.addsubstitutions(toxinidir=config.toxinidir, homedir=config.homedir)
         # As older versions of tox may have bugs or incompatibilities that
@@ -926,10 +986,9 @@ class parseini:
         else:
             config.toxworkdir = config.toxinidir.join(config.option.workdir, abs=True)
 
-        if not config.option.skip_missing_interpreters:
-            config.option.skip_missing_interpreters = reader.getbool(
-                "skip_missing_interpreters", False
-            )
+        if config.option.skip_missing_interpreters == "config":
+            val = reader.getbool("skip_missing_interpreters", False)
+            config.option.skip_missing_interpreters = "true" if val else "false"
 
         config.ignore_basepython_conflict = reader.getbool("ignore_basepython_conflict", False)
 
@@ -942,10 +1001,10 @@ class parseini:
 
         override = False
         if config.option.indexurl:
-            for urldef in config.option.indexurl:
-                m = re.match(r"\W*(\w+)=(\S+)", urldef)
+            for url_def in config.option.indexurl:
+                m = re.match(r"\W*(\w+)=(\S+)", url_def)
                 if m is None:
-                    url = urldef
+                    url = url_def
                     name = "default"
                 else:
                     name, url = m.groups()
@@ -966,7 +1025,7 @@ class parseini:
         self._make_thread_safe_path(config, "distdir", unique_id)
 
         reader.addsubstitutions(distdir=config.distdir)
-        config.distshare = reader.getpath("distshare", distshare_default)
+        config.distshare = reader.getpath("distshare", dist_share_default)
         self._make_thread_safe_path(config, "distshare", unique_id)
         reader.addsubstitutions(distshare=config.distshare)
         config.sdistsrc = reader.getpath("sdistsrc", None)
@@ -989,7 +1048,7 @@ class parseini:
 
         # configure testenvs
         for name in all_envs:
-            section = testenvprefix + name
+            section = "{}{}".format(testenvprefix, name)
             factors = set(name.split("-"))
             if (
                 section in self._cfg
@@ -1014,7 +1073,7 @@ class parseini:
             name = config.isolated_build_env
             if name not in config.envconfigs:
                 config.envconfigs[name] = self.make_envconfig(
-                    name, testenvprefix + name, reader._subs, config
+                    name, "{}{}".format(testenvprefix, name), reader._subs, config
                 )
 
     def _make_thread_safe_path(self, config, attr, unique_id):
@@ -1062,7 +1121,7 @@ class parseini:
             atype = env_attr.type
             try:
                 if atype in ("bool", "path", "string", "dict", "dict_setenv", "argv", "argvlist"):
-                    meth = getattr(reader, "get" + atype)
+                    meth = getattr(reader, "get{}".format(atype))
                     res = meth(env_attr.name, env_attr.default, replace=replace)
                 elif atype == "space-separated-list":
                     res = reader.getlist(env_attr.name, sep=" ")
@@ -1090,25 +1149,28 @@ class parseini:
         env_list = _split_env(env_str)
 
         # collect section envs
-        all_envs = set(env_list) - {"ALL"}
+        all_envs = OrderedDict((i, None) for i in env_list)
+        if "ALL" in all_envs:
+            all_envs.pop("ALL")
         for section in self._cfg:
             if section.name.startswith(testenvprefix):
-                all_envs.add(section.name[len(testenvprefix) :])
+                all_envs[section.name[len(testenvprefix) :]] = None
         if not all_envs:
-            all_envs.add("python")
+            all_envs["python"] = None
 
         package_env = config.isolated_build_env
         if config.isolated_build is True and package_env in all_envs:
-            all_envs.remove(package_env)
+            all_envs.pop(package_env)
 
         if not env_list or "ALL" in env_list:
-            env_list = sorted(all_envs)
+            env_list = list(all_envs.keys())
 
         if config.isolated_build is True and package_env in env_list:
             msg = "isolated_build_env {} cannot be part of envlist".format(package_env)
             raise tox.exception.ConfigError(msg)
 
-        return env_list, all_envs
+        all_env_list = list(all_envs.keys())
+        return env_list, all_env_list
 
 
 def _split_env(env):
@@ -1386,6 +1448,10 @@ class Replacer:
 
         if sub_type == "env":
             return self._replace_env(match)
+        if sub_type == "tty":
+            if is_interactive():
+                return match.group("substitution_value")
+            return match.group("default_value")
         if sub_type is not None:
             raise tox.exception.ConfigError(
                 "No support for the {} substitution type".format(sub_type)
@@ -1393,16 +1459,16 @@ class Replacer:
         return self._replace_substitution(match)
 
     def _replace_env(self, match):
-        envkey = match.group("substitution_value")
-        if not envkey:
+        key = match.group("substitution_value")
+        if not key:
             raise tox.exception.ConfigError("env: requires an environment variable name")
         default = match.group("default_value")
-        envvalue = self.reader.get_environ_value(envkey)
-        if envvalue is not None:
-            return envvalue
+        value = self.reader.get_environ_value(key)
+        if value is not None:
+            return value
         if default is not None:
             return default
-        raise tox.exception.MissingSubstitution(envkey)
+        raise tox.exception.MissingSubstitution(key)
 
     def _substitute_from_other_section(self, key):
         if key.startswith("[") and "]" in key:
@@ -1431,6 +1497,10 @@ class Replacer:
         return str(val)
 
 
+def is_interactive():
+    return sys.stdin.isatty()
+
+
 class _ArgvlistReader:
     @classmethod
     def getargvlist(cls, reader, value, replace=True):
@@ -1450,7 +1520,7 @@ class _ArgvlistReader:
             if not line:
                 continue
             if line.endswith("\\"):
-                current_command += " " + line[:-1]
+                current_command += " {}".format(line[:-1])
                 continue
             current_command += line
 
